/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include "rumd_data.h"

#include <unistd.h>

void CalcRDF(Conf *C0, long ***rdf, long ***sdf, int num_bins, double dx, unsigned int num_types) {
  unsigned int num_part = C0->num_part;

  double invDx = 1./dx;
  double    Lx = C0->metaData.L[0];
  double invLx = 1./Lx;
  double    Ly = C0->metaData.L[1];
  double invLy = 1./Ly;
  double    Lz = C0->metaData.L[2];
  double invLz = 1./Lz;
  int dr_index;

  for (unsigned int i=0; i<num_part; i++) {
    unsigned int TypeI =  C0->P[i].MyType;
    assert(TypeI<num_types);
    dr_index = (int)floor(( C0->P[i].x + Lx/2.)*invDx);
    if (dr_index < num_bins && dr_index >=0 ) sdf[TypeI][0][dr_index]++;
    dr_index = (int)floor(( C0->P[i].y + Ly/2.)*invDx);
    if (dr_index < num_bins && dr_index >=0 ) sdf[TypeI][1][dr_index]++;
    dr_index = (int)floor(( C0->P[i].z + Lz/2.)*invDx);
    if (dr_index < num_bins && dr_index >=0 ) sdf[TypeI][2][dr_index]++;

    for (unsigned int j=0; j<i; j++) {
      int TypeJ =  C0->P[j].MyType;
      double R = sqrt( C0->P[i].RsqMinImage(&C0->P[j], Lx, Ly, Lz, invLx, invLy, invLz ) );
      dr_index = (int)floor(R*invDx);
      if (dr_index < num_bins) {
	rdf[TypeI][TypeJ][dr_index]++;
             rdf[TypeJ][TypeI][dr_index]++;
      }
    }
  }
}

int main(int argc, char *argv[])
{

  unsigned int first_block=0;
  unsigned int last_block;

  int num_bins = 1000;   // No. bins in radial distribution function
  float min_dt = 0.0; // Only consider configurations spaced by at least this in time

  // assert(argc==2+1);
  // sscanf(argv[1], "%u", &first_block);
  // sscanf(argv[2], "%u", &last_block);
  // sscanf(argv[1], "%d", &num_bins);
  // sscanf(argv[2], "%f", &min_dt);

  static const char *usage = "Usage: %s [-n <number of bins>] [-m <minimum time between configurations>]\n";
  // FIXME fold extended argument description in here
  int opt;
  while ((opt = getopt(argc, argv, "hn:m:")) != -1) {
    switch (opt) {
    case 'n':
      num_bins = atoi(optarg); 
      break;
    case 'm':
      min_dt = strtof(optarg, NULL);
      break;
    case 'h':
      fprintf(stdout, usage, argv[0]);
      exit(EXIT_SUCCESS);
    case '?':
      fprintf(stderr, "%s: unknown option: -%c\n", argv[0], optopt);
      fprintf(stderr, usage, argv[0]);
      exit(EXIT_FAILURE);
      break;
    case ':':
      fprintf(stderr, "%s: missing option argument for option: -%c\n", argv[0], optopt);
      fprintf(stderr, usage, argv[0]);
      exit(EXIT_FAILURE);
      break;
    default:
      fprintf(stderr, usage, argv[0]);
      exit(EXIT_FAILURE);
    }
  }

  Conf *C0 = new Conf();

  gzFile gzfp;
  FILE *fp;

  fp=fopen("TrajectoryFiles/LastBlockSaved","r");
  if (!fp){
    fprintf(stderr, "%s: Error opening file TrajectoryFiles/LastBlockSaved: %s\n", argv[0], strerror(errno));
    exit(EXIT_FAILURE);
  }
  fscanf(fp, "%u", &last_block);
  fclose(fp);



  // Read metadata from first configuration in first block
  char *C0Name;
  asprintf(&C0Name,"./TrajectoryFiles/block%.4u.xyz.gz", first_block);
  printf("Reading block %u: %s\n\n", first_block, C0Name);
  gzfp=gzopen(C0Name,"r");
  if (!gzfp){
    fprintf(stderr, "%s: ReadConf: Error opening file %s: %s\n",argv[0], C0Name, strerror(errno));
    exit(EXIT_FAILURE);
  }
  C0->Read(gzfp,C0Name,1);
  gzclose(gzfp);
  free(C0Name);

  // Store the metaData we need
  float L  = C0->metaData.L[0];     // Assuming Lx=Ly=Lz for now...
  float dx = L/1.0/float(num_bins);
  float dt  = C0->metaData.dt;
  unsigned int num_part = C0->num_part;
  unsigned int num_types = C0->metaData.numTypes;
  unsigned int MaxIndex = C0->metaData.logLin.maxIndex;

  // Count and print the number of particles of each type
  int *NoPartOfType;
  NoPartOfType = new int[num_types];
  for (unsigned int i=0; i<num_types; i++) {
    NoPartOfType[i] = 0;
  }
  for (unsigned int i=0; i<num_part; i++) {
    unsigned int Type = C0->P[i].MyType;
    assert(Type<num_types);
    NoPartOfType[Type]++;
  }
  for (unsigned int i=0; i<num_types; i++) {
    printf("Particles of type %d: %d\n", i, NoPartOfType[i]);
  }

  // Allocate space for rdf's
  long ***rdf;
  rdf = new long**[num_types];
  for (unsigned int i=0; i<num_types; i++) {
     rdf[i] = new long*[num_types];
     for (unsigned int j=0; j<num_types; j++) {
        rdf[i][j] = new long[num_bins];
        for (int k=0; k<num_bins; k++) {
           rdf[i][j][k] = 0;
        }
     }
  }

  // Allocate space for spatial distributions
  long ***sdf;
  sdf = new long**[num_types];
  for (unsigned int i=0; i<num_types; i++) {
     sdf[i] = new long*[3];
     for (int j=0; j<3; j++) {
        sdf[i][j] = new long[num_bins];
        for (int k=0; k<num_bins; k++) {
           sdf[i][j][k] = 0;
        }
     }
  }


  int Count = 0;
  unsigned long int last_step_used;

  printf("\nReading block: ");
  for (unsigned int i=first_block; i<=last_block; i++) {

    // Read first configuration of block 'i'
    char *C0Name;
    asprintf(&C0Name,"./TrajectoryFiles/block%.4u.xyz.gz",i);
    if ((i-first_block)%10==0) printf("%u ", i);
    fflush(stdout);
    gzfp=gzopen(C0Name,"r");
    if (!gzfp){
      fprintf (stderr, "%s: ReadConf: Error opening file %s: %s\n",argv[0], C0Name, strerror(errno));
      exit(EXIT_FAILURE);
    }
    C0->Read(gzfp,C0Name,0);

    // Check that relevant metaData didn't change
    assert(C0->metaData.dt       == dt);
    assert(C0->num_part          == num_part);
    assert(C0->metaData.numTypes == num_types);
    assert(MaxIndex == C0->metaData.logLin.maxIndex);

    Count++;
    CalcRDF(C0, rdf, sdf, num_bins, dx,  num_types);
    last_step_used = C0->metaData.timeStepIndex;

    // Analyze configurations in the same block-file
    for (unsigned int j=1; j<=MaxIndex; j++) {
      C0->Read(gzfp,C0Name,0);

      if ( (C0->metaData.timeStepIndex - last_step_used)*dt > min_dt ) {
         Count++;
         CalcRDF(C0, rdf, sdf, num_bins, dx,  num_types);
	 last_step_used = C0->metaData.timeStepIndex;
      }
    }
    gzclose(gzfp);
    free(C0Name);
  }
  printf("\n");

  fp = fopen("rdf.dat","w");
  for (int k=0; k<num_bins; k++) {
     double r_start = k*dx;
     double r_last = (k+1)*dx;
     double rhoVol = 1./pow(L,3.) * 4./3.*M_PI*(pow(r_last,3.) - pow(r_start,3.));
     fprintf(fp, "%f ", (r_start + r_last)/2.);
     for (unsigned int i=0; i<num_types; i++ ) {
        for (unsigned int j=0; j<num_types; j++ ) {
           double factor = 1./float(Count)/rhoVol;
           if (i==j) factor /=float(NoPartOfType[i]*(NoPartOfType[i]-1));
           else factor /=float(NoPartOfType[i]*NoPartOfType[j]);
           fprintf(fp, "%f ", rdf[i][j][k]*factor);
        }
     }
     fprintf(fp, "\n");
  }
  fclose(fp);

  fp = fopen("sdf.dat","w");
  for (int k=0; k<num_bins; k++) {
     double r_start = k*dx;
     double r_last = (k+1)*dx;
     fprintf(fp, "%f ", (r_start + r_last)/2.);
     for (unsigned int i=0; i<num_types; i++ ) {
        for (int j=0; j<3; j++ ) {
           double factor = L/dx/float(Count)/float(num_part);
           fprintf(fp, "%f ", sdf[i][j][k]*factor);
        }
     }
     fprintf(fp, "\n");
  }
  fclose(fp);
}
