/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>
#include <string>
#include <getopt.h>

// includes, Class sample
#include "md_sample.h"
#include "md_IntegratorNoseHoover.h"
#include "md_timer.h"
#include "md_device.h"

int main( int argc, char* argv[]) {
  int parts_per_block = 16;
  int threads_per_part = 8;
  int thermostat = 0;                  // 0=none 1=Nose-Hoover
  int num_io_blocks = 1;               // Number of (logarthmic) configuration blocks
  int io_block_size = 1;
  int restartBlock = -1;               // which block to restart from (-1 means no restart)
  int dump_energies = 200;             // Interval (timesteps) between dumping of energies (0 means do not write energies)
  int momentum_reset_interval = 100;   // 0 to disable

  float Ti = 1.f;                      // Initial temperature
  float Tf = 1.f;                      // Final temperature
  float dt = 0.0025f;
  float skin = 0.3f;
  float Rscal = 1.f;                   // Scaling of initial configuration
  float relaxation = 0.3f;

  bool Tf_assigned = false;

  while (1) {
    static struct option long_options[] = {
      {"pb", 1, 0, 'p'},
      {"tp", 1, 0, 't'},
      {"thermostat", 1, 0, 'e'},
      {"numIOBlocks", 1, 0, 'o'},
      {"IOBlockSize", 1, 0, 'b'},
      {"dump_nrg", 1, 0, 'w'},
      {"restartBlock", 1, 0, 'x'},
      {"Ti", 1, 0, 'i'},
      {"Tf", 1, 0, 'f'},
      {"dt", 1, 0, 'd'},
      {"relaxation", 1, 0, 'u'},
      {"skin", 1, 0, 's'},
      {"Rscal", 1, 0, 'r'},
      {0, 0, 0, 0}
    };
    int option_index = 0;
    int c = getopt_long(argc, argv, "",long_options, &option_index);
    if (c == -1)
      break;

    switch (c) {
    case 0:
      printf("option %s", long_options[option_index].name);
      if (optarg)
        printf(" with arg %s", optarg);
        printf("\n");
      break;
    case 'p':
      parts_per_block = atoi(optarg);
      break;
    case 't':
      threads_per_part = atoi(optarg);
      break;
    case 'e':
      thermostat = atoi(optarg);
      break;
    case 'o':
      num_io_blocks = atoi(optarg);
      break;
    case 'b':
      io_block_size = atoi(optarg);
      break; 
    case 'w':
      dump_energies = atoi(optarg);
      break;
    case 'x':
      restartBlock = atoi(optarg);
      break;
    case 'i':
      Ti = strtof(optarg, 0);
      break;
    case 'f':
      Tf = strtof(optarg, 0);
      Tf_assigned = true;
      break;
    case 'd':
      dt = strtof(optarg, 0);
      break;
    case 'u':
      relaxation = strtof(optarg, 0);
      break;
    case 's':
      skin = strtof(optarg, 0);
      break;
    case 'r':
      Rscal = strtof(optarg, 0);
      break;
    case '?':
      exit(EXIT_FAILURE);
      break;
    default:
      printf("?? getopt returned character code 0%o ??\n", c);
    }
  }

  if (!Tf_assigned) Tf = Ti;

  md_device_init(argc, argv);

  // Create a new sample
  Sample* S = new Sample(parts_per_block, threads_per_part);

  // Create an integrator.
  IntegratorNoseHoover* itg =  new IntegratorNoseHoover(S);

  // Setup the integrator parameters.
  itg->SetTimeStep(dt);
  itg->SetRelaxationTime(relaxation);
  itg->SetTargetTemperature(Ti);
  S->SetIntegrator(itg);

  // Create LJ pair-potential
  Pot_LJ_12_6 potential;   
  
  // Setup parameters for Lennard-Jones interactions ( Kob & Andersen )
  // i  j   Sig   Eps  Rcut (in units of Sigma_ij)
  potential.SetParams(0, 0, 1.00, 1.00, 2.5);
  potential.SetParams(0, 1, 0.80, 1.50, 2.5);
  potential.SetParams(1, 0, 0.80, 1.50, 2.5);
  potential.SetParams(1, 1, 0.88, 0.50, 2.5);
  S->SetPotential(&potential, skin);
  
  // Read configuration & Copy it to device
  unsigned long int initialTimeStep = 0;
  
  if(restartBlock < 0) {
    S->ReadConf("start.xyz.gz");
    S->IsotropicScaleSystem(Rscal);
    itg->AllocateThermostatState();  
    itg->SetMomentumToZero();
  }  
  else{
    S->ReadRestartConf(restartBlock);
    itg->AllocateThermostatState();  
    initialTimeStep = (unsigned long int) (restartBlock+1)*io_block_size;
  }

  if(!thermostat) itg->SetThermostatState(0);

  S->DumpPotentials();
  S->UpdateNBlistInit();
  
  // Setup of various stuff
  S->SetIOBlockInfo(io_block_size);
  S->outputFormat.writeInterval = dump_energies;
  S->trajectoryFormat.has_pot_energies = true;
  S->trajectoryFormat.has_virials = true;
  
  if(num_io_blocks > 9999) {
    fprintf(stderr, "Error: Number of IO blocks cannot be more than 9999.\n");
    exit(1);
  }

  unsigned long int num_iter = (unsigned long int)num_io_blocks*io_block_size + 1;
  float dT = (Tf - Ti)/float(num_iter);
  
  printf("Total number of timestep(s) is %lu, ", num_iter);
  if(restartBlock >= 0)
    printf("restart: starting from iteration %lu\n",initialTimeStep);
  printf("dumping energies every %d timestep (every %f time-unit).\n", dump_energies, dump_energies*itg->GetTimeStep());
  
  // Execute the kernel(s)
  md_timer timer;
  
  for( unsigned long int i = initialTimeStep; i < num_iter; i++){
    S->UpdateNBlist();
    S->CalcF(&potential); // Calculate forces
    S->NotifyFileWriters(i,itg->GetTimeStep()); // Called here to ensure synchronization

    itg->Integrate(); // Integrate NVE or NVT dynamics.
    
    if(thermostat){
      itg->SetTargetTemperature(Ti + float(i)*dT);
      itg->UpdateThermostatState();
    }
    
    if ( momentum_reset_interval > 0 && i % momentum_reset_interval==0 ) 
      itg->SetMomentumToZero();
  } // end main loop
  
  double gpu_time = timer.elapsed();
  printf( "\nProcessing time GPU (s): %f\n", gpu_time);
  printf( " N  thermostat  pb  tp  ms/timestep: \n");
  float t_per_step = 0.;
  if (num_iter > 0) t_per_step = 1000.0*gpu_time/num_iter;
  printf( "%d %d   %d   %d  %f\n", S->num_part, thermostat, parts_per_block, threads_per_part, t_per_step);
  
  // Update forces and velocities and copy everything back from device 
  S->UpdateNBlistInit();
  S->CalcF(&potential);  
  S->WriteConf("confout.xyz");
  
  // Free objects.
  delete S; 
  delete itg;
}
