/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/
#include "md_device.h"

#include <cstdio>
#include <cstdlib>
#include <vector>

using namespace std;

void md_device_report()
{
  int dev;
  cudaGetDevice(&dev);
  cudaDeviceProp deviceProp;
  cudaError err = cudaGetDeviceProperties(&deviceProp, dev);
  if (err != cudaSuccess) {
    fprintf(stderr, "error: cudaGetDeviceProperties failed.\n");
    exit(EXIT_FAILURE);
  }
  fprintf(stdout, "CUDA: Using device %d: %s\n", dev, deviceProp.name);
}

void md_device_init(int argc, char *argv[])
{
  int deviceCount;
  cudaError err = cudaGetDeviceCount(&deviceCount);
  if (err != cudaSuccess) {
    fprintf(stderr, "error: cudaGetDeviceCount failed.\n");
    exit(EXIT_FAILURE); 
  }
  if (deviceCount == 0) {
    fprintf(stderr, "error: no devices supporting CUDA.\n");
    exit(EXIT_FAILURE);
  }
  if (deviceCount == 1) {
    // just one device or an emulated device present, no choice
    int dev=0;
    cudaDeviceProp deviceProp;
    cudaError err = cudaGetDeviceProperties(&deviceProp, dev);
    if (err != cudaSuccess) {
      fprintf(stderr, "error: cudaGetDeviceProperties failed.\n");
      exit(EXIT_FAILURE);
    }
    if (deviceProp.major < 1) {
        fprintf(stderr, "error: device does not support CUDA.\n");
        exit(EXIT_FAILURE);
    }
    cudaSetDevice(dev);
  }
  else {
    // several devices present, so make list of usable devices
    // and have one choosen among the currently available ones
    vector<int> usable_devices;
    for (int dev=0; dev<deviceCount; dev++) { 
      cudaDeviceProp deviceProp;
      cudaError err = cudaGetDeviceProperties(&deviceProp, dev);
      if (err != cudaSuccess) {
        fprintf(stderr, "error: cudaGetDeviceProperties failed.\n");
        exit(EXIT_FAILURE);
      }
      if ((deviceProp.major >= 1) && (deviceProp.multiProcessorCount >= 2) && (deviceProp.computeMode != cudaComputeModeProhibited)) {
        usable_devices.push_back(dev); 
      }
    }
    if (usable_devices.size() == 0) {
        fprintf(stderr, "error: no usable devices supporting CUDA.\n");
        exit(EXIT_FAILURE);
    }
    cudaError err = cudaSetValidDevices(&usable_devices[0], usable_devices.size());
    if (err != cudaSuccess ) {
      fprintf(stderr, "error: cudaSetValidDevices failed.\n");
      exit(EXIT_FAILURE);
    } 
    // trigger device initialization by a non-device management function call
    cudaThreadSynchronize();

    md_device_report();
  }
}
