/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/
#include "md_io_helpers.h"

#include <cstring>

void ConfMetaReader(char *commentLine, ConfMetaData *metaData, int verbose) {

  int nItems;
  unsigned int index;
  // for tokenizing: there is a two-level hierarchy of tokens,
  // (in some cases, eg list-arguments, three-level)
  // so we need to use strtok_r to allow nested tokenizing
  char *token;
  char *keyword;
  char *argument;
  char *listItem; // for list-arguments (comma-separated items)
  char *tokenMarker1; // highest level tokens
  char *tokenMarker2; // for separating keyword from argument
  char *tokenMarker3; // for list-arguments
	
  // possible entries in the columns list: 
  char requiredCols[][5] = {"type","x","y","z"};
  unsigned int numReq = sizeof(requiredCols)/(5*sizeof(char));
  char imStr[][5] = {"imx","imy","imz"};
  char velStr[][3] = {"vx","vy","vz"};
  char forStr[][3] = {"fx","fy","fz"};

  LogLin &log_lin = metaData->logLin;

  // top-level
  token = strtok_r(commentLine, " \t", &tokenMarker1);
  while(token) {
    // now token marks the next major token
    keyword = strtok_r(token, "=", &tokenMarker2);
    argument =  strtok_r(0, "=", &tokenMarker2);
    if(strcmp(keyword, "ioformat") == 0) {
      nItems = sscanf(argument, "%u", &(metaData->ioformat));
      if (verbose) printf("read ioformat %d\n",metaData->ioformat);
    }
    else if (strcmp(keyword, "timeStepIndex") == 0) {
      nItems = sscanf(argument, "%u", &(metaData->timeStepIndex));
      if (verbose) printf("read timeStepIndex %d\n",metaData->timeStepIndex);
    }
    else if (strcmp(keyword, "logLin") == 0) {
      listItem = strtok_r(argument, ",", &tokenMarker3); 
      nItems = sscanf(listItem, "%lu",&(log_lin.block));
      if (verbose) printf("read    log_lin.block %lu\n", log_lin.block);
      listItem = strtok_r(0, ",", &tokenMarker3); 
      nItems = sscanf(listItem, "%lu",&(log_lin.base));
      if (verbose) printf("read    log_lin.base %lu\n", log_lin.base);
      listItem = strtok_r(0, ",", &tokenMarker3);
      nItems = sscanf(listItem, "%lu",&(log_lin.index));
      if (verbose) printf("read    log_lin.index %lu\n", log_lin.index);
      listItem = strtok_r(0, ",", &tokenMarker3);
      nItems = sscanf(listItem, "%lu",&(log_lin.maxIndex));
      if (verbose) printf("read    log_lin.maxIndex %lu\n", log_lin.maxIndex);
      listItem = strtok_r(0, ",", &tokenMarker3); 
      nItems = sscanf(listItem, "%lu",&(log_lin.maxInterval));
      if (verbose) printf("read    log_lin.maxInterval %lu\n", log_lin.maxInterval);
      assert(0 == strtok_r(0, ",", &tokenMarker3));
      if (verbose)
	printf("read conf_index: Block %lu, Index %lu, MaxIndex %lu\n",
	       log_lin.block, log_lin.index, log_lin.maxIndex);
      metaData->has_logLin = true;
    }
    else if (strcmp(keyword, "numTypes") == 0) {
      nItems = sscanf(argument, "%u", &(metaData->numTypes));
      assert(metaData->numTypes <= MaxNumTypes);
      if (verbose) printf("read numTypes %d\n", metaData->numTypes);
    }
    else if (strcmp(keyword, "Nose-Hoover-Ps") == 0) {
      nItems = sscanf(argument, "%f", &(metaData->Ps));
      if (verbose) printf("Read Nose-Hoover-Ps %f\n", metaData->Ps);
    }
    else if (strcmp(keyword, "Barostat-Pv") == 0) {
      nItems = sscanf(argument, "%f", &(metaData->Pv));
      if (verbose) printf("read Barostat-Pv %f\n", metaData->Pv);
    }
    else if (strcmp(keyword, "dt") == 0) {
      nItems = sscanf(argument, "%f", &(metaData->dt));
      if (verbose) printf("read time step (dt) %f\n", metaData->dt);
    }		
    else if (strcmp(keyword, "mass") == 0) {
      listItem = strtok_r(argument, ",", &tokenMarker3);
      index = 0;
      while(listItem) {
	assert(index < metaData->numTypes);
	nItems = sscanf(listItem, "%f",&(metaData->massOfType[index]));
	index++;
	listItem = strtok_r(0, ",", &tokenMarker3);
      }
      
      if (verbose) printf("masses");
      for(index = 0; index < metaData->numTypes; ++index)
	if (verbose) printf(" %f", metaData->massOfType[index]);
      if (verbose) printf("\n");
    }
    else if (strcmp(keyword, "symbol") == 0) {
      listItem = strtok_r(argument, ",", &tokenMarker3);
      index = 0;
      while(listItem) {
	assert(index < metaData->numTypes);
	// not doing anything with symbol yet
	// (should include it in meta-data, but have to worry about
	// allocating space for potentially arbitrary length strings,
	// or make sure to read a maximum number of characters)
	if (verbose) printf("symbol %d is %s\n",index,listItem);
	index++;
	listItem = strtok_r(0, ",", &tokenMarker3);
      }
    }
    else if (strcmp(keyword, "boxLengths") == 0) {
      listItem = strtok_r(argument, ",", &tokenMarker3);
      unsigned int cd = 0;
      while(listItem) {
	assert(cd < DIM);
	nItems = sscanf(listItem, "%f",&(metaData->L[cd]));
	cd++;
	listItem = strtok_r(0, ",", &tokenMarker3);
      }
      if (verbose) printf("boxLengths");
      for(unsigned int cd = 0; cd < DIM; ++cd)
	if (verbose) printf(" %f", metaData->L[cd]);
      if (verbose) printf("\n");
    }
    else if (strcmp(keyword, "numMolecules") == 0) {
      
    }
    else if (strcmp(keyword, "columns") == 0) {
      listItem = strtok_r(argument, ",", &tokenMarker3);
      index = 0;
		
      
      while(listItem && index < numReq) {
	assert(strcmp(listItem, requiredCols[index]) == 0);
	 if (verbose) printf("columns, have required column %d (%s)\n",index,listItem);
	index++;
	listItem = strtok_r(0, ",", &tokenMarker3);
      }
      while(listItem) {
	// check if index < numReq (not all required columns present)?
	if(strcmp(listItem, "imx") == 0){
	  for(index = 0; index < DIM; ++index) {
	    assert(strcmp(listItem, imStr[index]) == 0);
	    if (verbose) printf("columns: have images: %s\n", listItem);
	    listItem = strtok_r(0, ",", &tokenMarker3);
	  }
	  metaData->has_images = true;
	}			
	else if(strcmp(listItem, "vx") == 0) {
	  for(index = 0; index < DIM; index++) {
	    assert(strcmp(listItem, velStr[index]) == 0);
	    if (verbose) printf("columns: have velocities %s\n", listItem);
	    listItem = strtok_r(0, ",", &tokenMarker3);
	  }
	  metaData->has_velocities = true;
	}
	else if(strcmp(listItem, "fx") == 0) {
	  for(index = 0; (index < DIM) ; index++) {
	    assert(strcmp(listItem, forStr[index]) == 0);
	    if (verbose) printf("columns: have forces %s\n", listItem);
	    listItem = strtok_r(0, ",", &tokenMarker3);
	  }
	  metaData->has_forces = true;
				}
	else if(strcmp(listItem, "pe") == 0) {
	  if (verbose) printf("columns: have pe\n");
	  metaData->has_pot_energies = true;
	  listItem = strtok_r(0, ",", &tokenMarker3);
	}
	else if(strcmp(listItem, "vir") == 0) {
	  metaData->has_virials = true;
	  if (verbose) printf("columns: have vir\n");
	  listItem = strtok_r(0, ",", &tokenMarker3);
				}
      } // end while (listItem) [columns]
    }
    else {
      fprintf(stderr, "ConfMetaReader: unknown keyword %s\n", keyword);
      exit(1);
    }
    
    // get the next major token
    token = strtok_r(0, " \t", &tokenMarker1);
  }
}



void EnergiesMetaReader(char *commentLine, OutputFormat *outputFormat, int verbose) {

  int nItems;
  // for tokenizing: there is a two-level hierarchy of tokens,
  // (in some cases, eg list-arguments, three-level)
  // so we need to use strtok_r to allow nested tokenizing
  char *token;
  char *keyword;
  char *argument;
  char *listItem; // for list-arguments (comma-separated items)
  char *tokenMarker1; // highest level tokens
  char *tokenMarker2; // for separating keyword from argument
  char *tokenMarker3; // for list-arguments

  // first set all options to false. If the appropriate string is present 
  // in columns, then the corrresponding option will be set to true
  std::map<std::string, bool>::const_iterator it;
  for(it = outputFormat->Start(); it != outputFormat->End(); it++ )
    outputFormat->Set(it->first, false);
	
  int ioformat = -1;
  unsigned int N = 0;
  float Dt = 0.;

  // read the "#" sign
  token = strtok_r(commentLine, " \t", &tokenMarker1);
  assert(strcmp(token,"#") == 0);
  // top-level
  token = strtok_r(0, " \t", &tokenMarker1);

    while(token) {
        
    // now token marks the next major token
    keyword = strtok_r(token, "=", &tokenMarker2);
    argument =  strtok_r(0, "=", &tokenMarker2);
    if(strcmp(keyword, "ioformat") == 0) {
      nItems = sscanf(argument, "%d", &ioformat);
      if (verbose) printf("read ioformat %d\n",ioformat);
    }
    else if(strcmp(keyword, "N") == 0) {
      nItems = sscanf(argument, "%u", &N);
      if (verbose) printf("read number of particles N = %d\n",N);
    }
    else if(strcmp(keyword, "Dt") == 0) {
      nItems = sscanf(argument, "%f", &Dt);
      if (verbose) printf("read Dt %f\n",Dt);
    }
    else if (strcmp(keyword, "columns") == 0) {
      listItem = strtok_r(argument, ",", &tokenMarker3);
      while(listItem) {
	for(it = outputFormat->Start(); it != outputFormat->End(); it++ )
	  if(strcmp(listItem, outputFormat->FileString(it->first).c_str()) == 0){
	    if (verbose) printf("columns: have %s\n", listItem);
	    assert(it->second == false); // (should not already be set)
	    outputFormat->Set(it->first, true);
	  } // if (strcmp ... )
	listItem = strtok_r(0, ",", &tokenMarker3);
      } // end while (listItem)
    } // end if "columns"
    else {
      fprintf(stderr, "EnergiesMetaReader: unknown keyword %s\n", keyword);
      exit(1);
      }
 
    // get the next major token
    token = strtok_r(0, " \t", &tokenMarker1);
  } // end while(token)
  
}
