#ifndef MD_SAMPLE_H
#define MD_SAMPLE_H

/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include "rumd_base.h"
#include "rumd_technical.h"

#include <cstdio>
#include <vector_types.h>
#include <cerrno>
// Following two lines to allow calling mkdir
#include <cstdlib>
#include <sys/stat.h>
#include <sys/types.h>
#include <zlib.h>
// Fixed placement.
#include "rumd_potentials.h"

#include "md_io_helpers.h"
#include "md_ParticleData.h"
#include "md_NeighborList.h"
#include "md_Integrator.h"

class Sample{

  // Integrators need to be able to modify particleData and neighborList. Friends are not inherited.
  friend class IntegratorNoseHoover; 
  friend class IntegratorNVU;

 protected:
  ParticleData particleData;
  NeighborList neighborList;
  Integrator* itg; 

 private:
  unsigned int maxNumIoBlocks;
  Potential* potential;
  
 public:
  unsigned int num_part;
  unsigned int numTypes;
  unsigned int num_blocks; 
  unsigned int num_virt_part; // No. 'particles' assigned threads (>=num_part)

  int particles_per_block;
  int threads_per_particle;
  int g; // Degrees of freedom.
  int allocatedNumBlocks; // For managing dynamic allocation.
  int allocatedNumParticles;

  float massOfType[MaxNumTypes];
  float skin;

  float*  h_Pv;    
  float*  h_L;
  float*  h_invL;
  float*  h_V;

  float*  d_Pv;
  float*  d_L;
  float*  d_invL;
  float*  d_V;

  float4* h_misc; 
  float4* d_misc; // Storage config temp.

  /////////////////////////////////////////////////////////
  ////////////// CUDA Parameters //////////////////////////
  /////////////////////////////////////////////////////////

  dim3 grid;
  dim3 threads;
  dim3 threads_tp1;
  size_t shared_size;
  size_t shared_size_tp1;

  /////////////////////////////////////////////////////////
  ////////////// I/O Vars /////////////////////////////////
  /////////////////////////////////////////////////////////

  std::string trajectoryDir;
  gzFile gz_energyFile;

  int ioformat;  
  bool writers_initialized;
  bool writeTrajectoryConfs;
  bool newEnergyFile;

  ConfMetaData restartFormat;
  ConfMetaData trajectoryFormat;
  OutputFormat outputFormat; 
  
  /////////////////////////////////////////////////////////
  ////////////// Constructors/Methods /////////////////////
  /////////////////////////////////////////////////////////

  Sample( unsigned int set_particles_per_block, unsigned int set_threads_per_particle );
  ~Sample();
  
  /////////////////////////////////////////////////////////
  ////////////// Allocation ///////////////////////////////
  /////////////////////////////////////////////////////////
  
  void AllocateBlocks(int nb);
  void AllocateParticles(int nvp);
  void AllocateBlocksParticles(int nb_vp);
  void FreeBlocks();
  void FreeParticles();
  void FreeBlocksParticles();
  void FreeAll();

  /////////////////////////////////////////////////////////
  ////////////// I/O //////////////////////////////////////
  /////////////////////////////////////////////////////////

  void InitializeFileWriters(unsigned long int step);
  void NotifyFileWriters(unsigned long int step, float dt);
  void ReadConf( const std::string& filename );
  void ReadRestartConf(int restartBlock);
  void WriteConf( const std::string&filename, ConfMetaData *metaData=0 );
  void WriteEnergies(float dt);
  void CalcEnergies(float4* sumU, float4* sumK, float4* sumW, float4* sumS, float4* sumC, float *stress);
  void SetIOBlockInfo(unsigned int blockSize, unsigned int base=1, unsigned int maxInterval=0);
  void SetNumberOfParticles(int num_part);
  void InitializeBoxVariables(float L); 
  void SetPotential(Potential* pot, double skin);  
  void DumpPotentials();
  void CreateRandomConf(unsigned int num_part, float L, float frac_A, float T);
  float4 CalculateCM( float4* pos_array, float4* velo_array, int index );
  void IsotropicScaleSystem( float Rscal );
  void ZeroAllVelocities();

  /////////////////////////////////////////////////////////
  ////////////// Force/Nblist/Constraints /////////////////
  /////////////////////////////////////////////////////////

  void UpdateNBlist();
  void UpdateNBlistInit();
  void CalcF(Potential *P, bool calc_stresses = false, bool calc_temp = false, bool calc_bs = false);

  /////////////////////////////////////////////////////////
  ////////////// CopyFrom Methods /////////////////////////
  /////////////////////////////////////////////////////////

  void CopyDataFromDevice();
  void CopyMiscFromDevice();
  void CopyLFromDevice();
  void CopyinvLFromDevice();
  void CopyVolumeFromDevice();

  /////////////////////////////////////////////////////////
  ////////////// CopyTo Methods ///////////////////////////
  /////////////////////////////////////////////////////////

  void CopyConfToDevice();
  void CopyBoxToDevice();
  
  // Access methods.
  const ParticleData& GetParticleData() const { return particleData; }
  const NeighborList& GetNeighborList() const { return neighborList; }

  // Set the current integrator.
  void SetIntegrator(Integrator* newItg){ itg = newItg; } 
};

#endif // MD_SAMPLE_H
